SERVICE_MARKETS_MESSAGE_CHANNEL = "SERVICE_MARKETS_TEST_V0.4"
SERVICE_MARKETS_MANAGER_PUBKEYS = [
    "Bxa95pz5SkcKQE5Qji893inhxeXsnhVK8uALoF993fVv",  # Mike
    "6ZSRBF5tPYZg99mqp99w11tiePaaUn6HSZH1GKCn2w23",  # Mike Docker
    "82AocbGgNL8uHeVqMiyL1phUdeyxW6ktTpfQk6syCcgF",  # Leonardo
    "7TTtzzYqWXtTcAqrsRzHPWa6T3fnkW6GjdkNe58TLYGJ",  # Leonardo Docker
    "6Bf4i3vGhjtz69y2bw1tpYPmCGR4PmgwW8JLTF2qy4Z2",  # Leonardo Extra
    "BExGtmNQP4eQjSNNuhWiGnUGW6aucZeJB7d3uAaATcc5",  # Ricardo
]

API_MESSAGE_FILTER = [
    {
        "channel": SERVICE_MARKETS_MESSAGE_CHANNEL,
        "type": "POST",
        "post_type": [
            "Service",
            "UserInfo",
            "Vote",
            "Comment",
            "Permission",
            "amend",
        ],
    }
]



from enum import Enum
from typing import List, Optional

from aars import Record
from pydantic import Field


class UserInfo(Record):
    username: str
    address: str
    bio: Optional[str]
    email: Optional[str]
    link: Optional[str]


class Votable(Record):
    upvotes: int = 0
    downvotes: int = 0


class Service(Votable):
    name: str
    description: str
    url: str
    image_url: Optional[str]
    price: float
    tags: List[str]
    owner_address: str
    comment_counter: int = 0
    payment_id: Optional[str] = None


class VoteType(str, Enum):
    UP = "up"
    DOWN = "down"


class VotableType(str, Enum):
    SERVICE = "service"
    COMMENT = "comment"


class Vote(Record):
    item_id: str
    item_type: VotableType
    user_address: str
    vote: VoteType


class Comment(Votable):
    service_id: str
    user_address: str
    comment: str


class Permission(Record):
    user_address: str
    service_id: str


class Payment(Record):
    contractAddress: str
    tokenAddress: str
    txHash: str
    to: str
    from_: str = Field(alias='from')
    amount: str
    reference: str



# Desc: FastAPI Depends plugin for controlling access to API endpoints
# It will raise a 403 if the user is not allowed to access the endpoint.
# For the first request, the Aleph network will be queried to see if the user is allowed to access the endpoint.
import asyncio
from typing import Optional

from aars import AARS
from aleph.sdk import AuthenticatedAlephClient
from aleph.sdk.chains.sol import get_fallback_account
from aleph.sdk.conf import settings
from fastapi import HTTPException, FastAPI
from fastapi_walletauth import WalletAuth
from fastapi_walletauth.core import SignatureChallengeTokenAuth
from starlette.requests import Request
from starlette.middleware.base import BaseHTTPMiddleware

from .model import Permission, Service
from .session import initialize_aars


class ServicePermissionAuth(SignatureChallengeTokenAuth):
    aars: AARS
    service_record: Optional[Service] = None
    cached_permissions = {}

    def __init__(
        self,
        service_url: str,
    ):
        super().__init__()
        self.service_url = service_url

    def __call__(self, request: Request) -> WalletAuth:
        """
        Check if the user has the given permission for the given service.
        """
        wallet_auth: WalletAuth = super().__call__(request)
        if self.cached_permissions.get(wallet_auth.address):
            return wallet_auth
        loop = self.aars.session.http_session.loop
        permission_record = loop.run_until_complete(Permission.filter(
            user_address=wallet_auth.address,
            service_id=self.service_record.item_hash,
        ).all())
        if not permission_record:
            raise HTTPException(
                status_code=403,
                detail="User does not have permission to access this service",
            )
        self.cached_permissions[wallet_auth.address] = permission_record
        return wallet_auth

    async def setup(self, **kwargs):
        self.aars = await initialize_aars(**kwargs)
        service = await Service.filter(url=self.service_url).first()
        if not service:
            raise ValueError(
                f"Service with url {self.service_url} is not registered on service.markets"
            )
        print(f"Service {self.service_url} successfully loaded. Heimdall is ready.")
        self.service_record = service


class HeimdallMiddleware(BaseHTTPMiddleware):
    def __init__(self, app: FastAPI, backend: ServicePermissionAuth, **kwargs):
        super().__init__(app)
        self.backend = backend
        self.kwargs = kwargs
        # setup backend
        account = kwargs.get("account", get_fallback_account())
        aleph_session = AuthenticatedAlephClient(account, settings.API_HOST)
        loop = aleph_session.http_session.loop
        loop.run_until_complete(self.backend.setup(**kwargs))

    async def dispatch(self, request, call_next):
        if not request.url.path.startswith("/authorization"):
            request.state.wallet_auth = self.backend(request)
        return await call_next(request)


def setup_heimdall(app: FastAPI, service_url: str, **kwargs):
    """
    Setup Heimdall middleware for the given app. This will check if the user has permission to access the given service.
    Permission is checked by querying the Aleph network.
    """
    app.add_middleware(
        HeimdallMiddleware,
        backend=ServicePermissionAuth(service_url),
        backend_kwargs=kwargs,
    )



import asyncio
from typing import List, Optional, TypeVar, Tuple

from fastapi import APIRouter, HTTPException
from fastapi_walletauth import WalletAuthDep

from ...core.model import (
    Service,
    Permission,
    Comment,
    VoteType,
    Vote,
    VotableType, Payment,
)
from ..api_model import (
    ServiceWithPermissionStatus,
    UploadServiceRequest,
    VoteServiceResponse,
    VoteCommentResponse, PutInvoiceServiceResponse,
)
from ...core.request_network import fetch_payment

router = APIRouter(
    prefix="/services",
    tags=["services"],
    responses={404: {"description": "Not found"}},
)

T = TypeVar("T", Service, Comment)


@router.get("")
async def get_services(
    view_as: Optional[str] = None,
    by: Optional[str] = None,
    page: int = 1,
    page_size: int = 20,
) -> List[ServiceWithPermissionStatus]:
    """
    Get all services or filter by owner address. Use `view_as` to get the permission status for a given user.
    """
    services: List[Service] = []
    if by:
        services = await Service.filter(owner_address=by).page(
            page=page, page_size=page_size
        )
    else:
        services = await Service.fetch_objects().page(page=page, page_size=page_size)

    services_response: List[ServiceWithPermissionStatus] = []
    if view_as:
        permissions = await Permission.filter(user_address=view_as).all()
        for service in services:
            permission = filter(lambda p: p.serviceID == service.item_hash, permissions)
            services_response.append(
                ServiceWithPermissionStatus(
                    **service.dict(), permission_status=permission is not None
                )
            )
    else:
        services_response = [
            ServiceWithPermissionStatus(**service.dict(), permission_status=None)
            for service in services
        ]
    return services_response


@router.put("")
async def upload_service(
    service: UploadServiceRequest, wallet: WalletAuthDep
) -> Service:
    """
    Upload a service.
    If an `item_hash` is provided, it will update the service with that id.
    """
    if service.owner_address != wallet.address:
        raise HTTPException(
            status_code=403,
            detail="address does not match currently authorized user wallet",
        )
    if service.item_hash is not None:
        old_service = await Service.fetch(service.item_hash).first()
        if old_service:
            old_service.name = service.name
            old_service.description = service.description
            old_service.owner_address = service.owner_address
            old_service.item_hash = service.item_hash
            old_service.url = service.url
            old_service.image_url = service.image_url
            old_service.tags = service.tags
            old_service.price = service.price
        else:
            raise HTTPException(status_code=404, detail="No Service found")
    return await Service(**service.dict()).save()


@router.get("/{service_id}")
async def get_service(
    service_id: str, view_as: Optional[str] = None
) -> ServiceWithPermissionStatus:
    """
    Get a specific service by id.
    """
    service = await Service.fetch(service_id).first()
    if not service:
        raise HTTPException(status_code=404, detail="No Service found")
    if view_as:
        permission = await Permission.filter(
            user_address=view_as, service_id=service_id
        ).first()
        return ServiceWithPermissionStatus(
            **service.dict(), permission_status=permission is not None
        )
    return ServiceWithPermissionStatus(**service.dict(), permission_status=None)


@router.get("/{service_id}/permissions")
async def get_service_permissions(service_id: str) -> List[Permission]:
    """
    Get all granted permissions for a given service.
    """
    permissions = await Permission.filter(service_id=service_id).all()
    return permissions


@router.put("/{service_id}/vote")
async def vote_service(
    service_id: str,
    vote: VoteType,
    wallet: WalletAuthDep,
) -> VoteServiceResponse:
    """
    Update your vote for a given service.
    """
    service = await Service.fetch(service_id).first()
    if not service:
        raise HTTPException(status_code=404, detail="No Service found")
    vote_record = await Vote.filter(
        item_id=service_id, user_address=wallet.address
    ).first()
    if not vote_record:
        vote_record = Vote(
            item_id=service_id,
            item_type=VotableType.SERVICE.value,
            user_address=wallet.address,
            vote=vote,
        )
    else:
        vote_record.vote = vote
    service, vote_record = await update_vote(service, vote_record)
    return VoteServiceResponse(service=service, vote=vote_record)


@router.put("/{service_id}/payment/{tx_hash}")
async def put_invoice_service(
    service_id: str,
    tx_hash: str,
    wallet: WalletAuthDep,
) -> PutInvoiceServiceResponse:
    """
    Update your payment for a given service.
    """
    service = await Service.fetch(service_id).first()
    if not service:
        raise HTTPException(status_code=404, detail="No Service found")
    payment = await Payment.filter(txHash=tx_hash).first()
    if payment:
        raise HTTPException(status_code=409, detail="Payment already registered")
    payment = await fetch_payment(tx_hash)
    if not payment:
        raise HTTPException(status_code=404, detail="No Payment found")
    if payment.from_ != wallet.address:
        raise HTTPException(
            status_code=403,
            detail="payment does not match currently authorized user wallet",
        )
    await payment.save()
    service.payment_id = payment.item_hash
    permission = Permission(
        service_id=service_id,
        user_address=wallet.address,
    )
    service, permission = await asyncio.gather(
        service.save(),
        permission.save(),
    )
    return PutInvoiceServiceResponse(service=service, permission=permission, payment=payment)


@router.get("/{service_id}/comments")
async def get_service_comments(
    service_id: str, page: int = 1, page_size: int = 20
) -> List[Comment]:
    """
    Get all comments for a given service.
    """
    comments = await Comment.filter(service_id=service_id).page(
        page=page, page_size=page_size
    )
    return comments


@router.post("/{service_id}/comments")
async def post_service_comment(
    service_id: str,
    comment: str,
    wallet: WalletAuthDep,
) -> Comment:
    """
    Post a comment for a given service.
    """
    return await Comment(
        service_id=service_id,
        comment=comment,
        user_address=wallet.address,
    ).save()


@router.put("/{service_id}/comments/{comment_id}/vote")
async def vote_service_comment(
    service_id: str,
    comment_id: str,
    vote: VoteType,
    wallet: WalletAuthDep,
) -> VoteCommentResponse:
    """
    Update your vote for a given comment.
    """
    comment = await Comment.fetch(comment_id).first()
    if not comment:
        raise HTTPException(status_code=404, detail="No Comment found")
    vote_record = await Vote.filter(
        item_id=comment_id, user_address=wallet.address
    ).first()
    if not vote_record:
        vote_record = Vote(
            item_id=comment_id,
            item_type=VotableType.COMMENT.value,
            user_address=wallet.address,
            vote=vote,
        )
    else:
        vote_record.vote = vote
    comment, vote_record = await update_vote(comment, vote_record)
    return VoteCommentResponse(comment=comment, vote=vote_record)


async def update_vote(votable: T, vote_record: Vote) -> Tuple[T, Vote]:
    print(vote_record.vote.value, VoteType.UP.value)
    if vote_record.item_hash is not None:
        if vote_record.vote.value == VoteType.UP.value and vote_record.changed:
            votable.downvotes -= 1
            votable.upvotes += 1
        elif vote_record.vote.value == VoteType.DOWN.value and vote_record.changed:
            votable.downvotes += 1
            votable.upvotes -= 1
        await asyncio.gather(
            votable.save(),
            vote_record.save(),
        )
    else:
        votable.upvotes += 1 if vote_record.vote.value == VoteType.UP.value else 0
        votable.downvotes += 1 if vote_record.vote.value == VoteType.DOWN.value else 0
        vote_record, votable = await asyncio.gather(
            vote_record.save(),
            votable.save(),
        )
    return votable, vote_record



import asyncio
import logging
import os
from os import listdir
from typing import Optional

from aars import AARS, Record
from aleph.sdk.vm.app import AlephApp
from aleph_message.models import PostMessage
from fastapi import FastAPI
from fastapi.middleware.cors import CORSMiddleware
from fastapi_walletauth import authorization_routes

from ..core.constants import API_MESSAGE_FILTER, SERVICE_MARKETS_MESSAGE_CHANNEL
from ..core.session import initialize_aars
from .routers import (
    services,
    users,
)

logger = (
    logging.getLogger(__name__)
    if __name__ != "__main__"
    else logging.getLogger("uvicorn")
)
http_app = FastAPI()

origins = ["*"]

http_app.add_middleware(
    CORSMiddleware,
    allow_origins=origins,
    allow_credentials=True,
    allow_methods=["*"],
    allow_headers=["*"],
)

http_app.include_router(services.router)
http_app.include_router(users.router)
http_app.include_router(authorization_routes)

app = AlephApp(http_app=http_app)


async def re_index():
    logger.info(f"API re-indexing channel {AARS.channel}")
    await asyncio.wait_for(AARS.sync_indices(), timeout=None)
    logger.info("API re-indexing done")


@app.on_event("startup")
async def startup():
    app.aars = await initialize_aars()
    print("Syncing indices...")
    await re_index()


@app.get("/")
async def index():
    if os.path.exists("/opt/venv"):
        opt_venv = list(listdir("/opt/venv"))
    else:
        opt_venv = []
    return {
        "vm_name": SERVICE_MARKETS_MESSAGE_CHANNEL,
        "endpoints": [
            "/docs",
        ],
        "files_in_volumes": {
            "/opt/venv": opt_venv,
        },
    }


@app.post("/event")
async def event(event: PostMessage):
    await fishnet_event(event)


@app.get("/address")
async def address():
    return AARS.account.get_address()


@app.event(filters=API_MESSAGE_FILTER)
async def fishnet_event(event: PostMessage):
    record: Optional[Record]
    print("fishnet_event", event)
    api_message_filter_without_amend = API_MESSAGE_FILTER.copy()
    api_message_filter_without_amend.remove("amend")
    if event.content.type in api_message_filter_without_amend:
        if Record.is_indexed(event.item_hash):
            return
        cls: Record = globals()[event.content.type]
        record = await cls.from_post(event)
    else:  # amend
        if Record.is_indexed(event.content.ref):
            return
        record = await Record.fetch(event.content.ref).first()
    assert record
    for inx in record.get_indices():
        inx.add_record(record)
